"""Top-level package for hagerstrand."""

__author__ = """Jimmy Feng"""
__email__ = 'jfeng1996@gmail.com'
__version__ = '0.0.11'

from .hagerstrand import *
from .dataprocess import *
from .common import *
from .utils import *


"""Main module for the hagerstrand package."""
import os
import ipyleaflet
import ee
import box
#import math
from ipyleaflet import FullScreenControl, LayersControl, DrawControl, MeasureControl, ScaleControl, TileLayer, basemaps, basemap_to_tiles
#import ipywidgets
from sklearn.neighbors import BallTree
import numpy as np
import pandas as pd
import geopandas as gpd
import json
from .utils import random_string
from .common import ee_initialize, tool_template
from .toolbar import main_toolbar, filter_df_widget


# Credit: Dr. Qiusheng Wu
class Map(ipyleaflet.Map):
    """This Map class inherits the ipyleaflet Map class.

    Args:
        ipyleaflet (ipyleaflet.Map()): An ipyleaflet map.
    """    
    def __init__(self, **kwargs):
        
        if "center" not in kwargs:
            kwargs["center"] = [40, -100]

        if "zoom" not in kwargs:
            kwargs["zoom"] = 4

        if "scroll_wheel_zoom" not in kwargs:
            kwargs["scroll_wheel_zoom"] = True

        super().__init__(**kwargs)

        if "height" not in kwargs:
            self.layout.height = "600px"
        else:
            self.layout.height = kwargs["height"]

        self.add_control(FullScreenControl())
        self.add_control(LayersControl(position="topright"))
        self.add_control(DrawControl(position="topleft"))
        self.add_control(MeasureControl())
        self.add_control(ScaleControl(position="bottomleft"))
        self.toolbar = None

        # tools = {
        #     "folder-open": {
        #         "name": "open_data",
        #         "tooltip": "Open local vector/raster data"
        #     },
        #     "map": {
        #         "name": "basemap",
        #         "tooltip": "Change basemap"
        #     },
        #     "gears": {
        #         "name": "whitebox",
        #         "tooltip": "WhiteboxTools for local geoprocessing"
        #     },
        #     "filter": {
        #         "name": "query",
        #         "tooltip": "Attribute selection"
        #     },
        # }
        # icons = list(tools.keys())
        # tooltips = [item["tooltip"] for item in list(tools.values())]

        # icon_width = "32px"
        # icon_height = "32px"
        # n_cols = 3
        # n_rows = math.ceil(len(icons) / n_cols)

        # toolbar_grid = ipywidgets.GridBox(
        #     children=[
        #         ipywidgets.ToggleButton(
        #             layout=ipywidgets.Layout(
        #                 width="auto", height="auto", padding="0px 0px 0px 4px"
        #             ),
        #             button_style="primary",
        #             icon=icons[i],
        #             tooltip=tooltips[i],
        #         )
        #         for i in range(len(icons))
        #     ],
        #     layout=ipywidgets.Layout(
        #         width="107px",
        #         grid_template_columns=(icon_width + " ") * n_cols,
        #         grid_template_rows=(icon_height + " ") * n_rows,
        #         grid_gap="1px 1px",
        #         padding="5px",
        #     ),
        # )
        # self.toolbar = toolbar_grid


        if "google_map" not in kwargs:
            layer = TileLayer(
                url="https://mt1.google.com/vt/lyrs=m&x={x}&y={y}&z={z}",
                attribution="Google",
                name="Google Maps",
            )
            self.add_layer(layer)
        else:
            if kwargs["google_map"] == "ROADMAP":
                layer = TileLayer(
                    url="https://mt1.google.com/vt/lyrs=m&x={x}&y={y}&z={z}",
                    attribution="Google",
                    name="Google Maps",
                )
                self.add_layer(layer)
            elif kwargs["google_map"] == "HYBRID":
                layer = TileLayer(
                    url="https://mt1.google.com/vt/lyrs=y&x={x}&y={y}&z={z}",
                    attribution="Google",
                    name="Google Satellite"
                )
                self.add_layer(layer)

        if "basemap" not in kwargs:
            layer = basemap_to_tiles(basemaps.OpenStreetMap.Mapnik)
            self.add_layer(layer)
        else:
            layer = basemap_to_tiles(kwargs["basemap"])
            self.add_layer(layer)

        main_toolbar(self)

    def add_geojson(self, in_geojson, style=None, layer_name="Untitled"):
        """Adds a GeoJSON file to the map.

        Args:
            in_geojson (str): The file path to the input GeoJSON.
            style (dict, optional): The style for the GeoJSON layer. Defaults to None.
            layer_name (str, optional): The layer name for the GeoJSON layer. Defaults to "Untitled".

        Raises:
            FileNotFoundError: If the provided file path does not exist.
            TypeError: If the input GeoJSON is not a str.
        """        
        import json

        if isinstance(in_geojson, str):

            if not os.path.exists(in_geojson):
                raise FileNotFoundError("The provided GeoJSON file could not be found.")

            with open(in_geojson) as f:
                data = json.load(f)
        
        elif isinstance(in_geojson, dict):
            data = in_geojson
        
        else:
            raise TypeError("The input geojson must be a type of str or dict.")

        if style is None:
            style = {
                "stroke": True,
                "color": "#000000",
                "weight": 2,
                "opacity": 1,
                "fill": True,
                "fillColor": "#0000ff",
                "fillOpacity": 0.4,
            }

        geo_json = ipyleaflet.GeoJSON(data=data, style=style, name=layer_name)
        self.add_layer(geo_json) 

# Credit: Dr. Qiusheng Wu
    def add_shapefile(self, in_shp, style=None, layer_name="Untitled"):
        """Adds a shapefile to the map

        Args:
            in_shp (str): The file path to the input shapefile.
            style (dict, optional): The style for the shapefile. Defaults to None.
            layer_name (str, optional): The layer name for the shapefile layer. Defaults to "Untitled".
        """
        geojson = shp_to_geojson(in_shp)
        self.add_geojson(geojson, style=style, layer_name=layer_name)

    def add_gmapjson(self, in_json, style=None, layer_name="Untitled"):
        """Adds a Google Map Location History JSON file to the map

        Args:
            in_json (str): The file path to the input JSON.
            style (dict, optional): The style for the JSON. Defaults to None.
            layer_name (str, optional): The layer name for the JSON layer. Defaults to "Untitled".
        """
        geojson = gmapjson_to_geojson(in_json)
        self.add_geojson(geojson, style=style, layer_name=layer_name)

    def add_gdf(self, in_gdf, style=None, layer_name="Untitled"):
        """Adds a Pandas DataFrame to the map

        Args:
            in_gdf (gpd.GeoDataFrame): The GeoPandas GeoDataFrame.
            style (dict, optional): The style for the GeoDataFrame. Defaults to None.
            layer_name (str, optional): The layer name for the GeoDataFrame layer. Defaults to "Untitled".
        """
        geodata = ipyleaflet.GeoData(geo_dataframe = in_gdf, name=layer_name)
        self.add_layer(geodata)

    # Source: Dr. Qiusheng Wu: https://github.com/giswqs/geemap/blob/master/geemap/geemap.py
    def set_center(self, lon, lat, zoom=None):
        """Centers the map view at a given coordinates with the given zoom level. Source: Dr. Qiusheng Wu: https://github.com/giswqs/geemap/blob/master/geemap/geemap.py
        Args:
            lon (float): The longitude of the center, in degrees.
            lat (float): The latitude of the center, in degrees.
            zoom (int, optional): The zoom level, from 1 to 24. Defaults to None.
        """
        self.center = (lat, lon)
        if zoom is not None:
            self.zoom = zoom

    setCenter = set_center

    # Source: Dr. Qiusheng Wu: https://github.com/giswqs/geemap/blob/master/geemap/geemap.py
    def center_object(self, ee_object, zoom=None):
        """Centers the map view on a given object. Source: Dr. Qiusheng Wu: https://github.com/giswqs/geemap/blob/master/geemap/geemap.py
        Args:
            ee_object (Element|Geometry): An Earth Engine object to center on a geometry, image or feature.
            zoom (int, optional): The zoom level, from 1 to 24. Defaults to None.
        """
        if zoom is None and hasattr(self, "fit_bounds"):
            self.zoom_to_object(ee_object)
        else:
            lat = 0
            lon = 0
            if isinstance(ee_object, ee.geometry.Geometry):
                centroid = ee_object.centroid(1)
                lon, lat = centroid.getInfo()["coordinates"]
            else:
                try:
                    centroid = ee_object.geometry().centroid(1)
                    lon, lat = centroid.getInfo()["coordinates"]
                except Exception as e:
                    print(e)
                    raise Exception(e)

            self.setCenter(lon, lat, zoom)

    centerObject = center_object

    # Source: Dr. Qiusheng Wu: https://github.com/giswqs/geemap/blob/master/geemap/geemap.py

    def add_ee_layer(
        self, ee_object, vis_params={}, name=None, shown=True, opacity=1.0
    ):
        """Adds a given EE object to the map as a layer. Source: Dr. Qiusheng Wu: https://github.com/giswqs/geemap/blob/master/geemap/geemap.py
        Args:
            ee_object (Collection|Feature|Image|MapId): The object to add to the map.
            vis_params (dict, optional): The visualization parameters. Defaults to {}.
            name (str, optional): The name of the layer. Defaults to 'Layer N'.
            shown (bool, optional): A flag indicating whether the layer should be on by default. Defaults to True.
            opacity (float, optional): The layer's opacity represented as a number between 0 and 1. Defaults to 1.
        """
        from box import Box

        image = None
        if name is None:
            layer_count = len(self.layers)
            name = "Layer " + str(layer_count + 1)

        if (
            not isinstance(ee_object, ee.Image)
            and not isinstance(ee_object, ee.ImageCollection)
            and not isinstance(ee_object, ee.FeatureCollection)
            and not isinstance(ee_object, ee.Feature)
            and not isinstance(ee_object, ee.Geometry)
        ):
            err_str = "\n\nThe image argument in 'addLayer' function must be an instace of one of ee.Image, ee.Geometry, ee.Feature or ee.FeatureCollection."
            raise AttributeError(err_str)

        if (
            isinstance(ee_object, ee.geometry.Geometry)
            or isinstance(ee_object, ee.feature.Feature)
            or isinstance(ee_object, ee.featurecollection.FeatureCollection)
        ):
            features = ee.FeatureCollection(ee_object)

            width = 2

            if "width" in vis_params:
                width = vis_params["width"]

            color = "000000"

            if "color" in vis_params:
                color = vis_params["color"]

            image_fill = features.style(**{"fillColor": color}).updateMask(
                ee.Image.constant(0.5)
            )
            image_outline = features.style(
                **{"color": color, "fillColor": "00000000", "width": width}
            )

            image = image_fill.blend(image_outline)
        elif isinstance(ee_object, ee.image.Image):
            image = ee_object
        elif isinstance(ee_object, ee.imagecollection.ImageCollection):
            image = ee_object.mosaic()

        if "palette" in vis_params and isinstance(vis_params["palette"], Box):
            try:
                vis_params["palette"] = vis_params["palette"]["default"]
            except Exception as e:
                print("The provided palette is invalid.")
                raise Exception(e)

        map_id_dict = ee.Image(image).getMapId(vis_params)
        tile_layer = TileLayer(
            url=map_id_dict["tile_fetcher"].url_format,
            attribution="Google Earth Engine",
            name=name,
            opacity=opacity,
            visible=shown,
        )

        layer = self.find_layer(name=name)
        if layer is not None:

            existing_object = self.ee_layer_dict[name]["ee_object"]

            if isinstance(existing_object, ee.Image) or isinstance(
                existing_object, ee.ImageCollection
            ):
                self.ee_raster_layers.remove(existing_object)
                self.ee_raster_layer_names.remove(name)
                if self.plot_dropdown_widget is not None:
                    self.plot_dropdown_widget.options = list(self.ee_raster_layer_names)
            elif (
                isinstance(ee_object, ee.Geometry)
                or isinstance(ee_object, ee.FeatureCollection)
            ):
                self.ee_vector_layers.remove(existing_object)
                self.ee_vector_layer_names.remove(name)

            self.ee_layers.remove(existing_object)
            self.ee_layer_names.remove(name)
            self.remove_layer(layer)

        self.ee_layers.append(ee_object)
        if name not in self.ee_layer_names:
            self.ee_layer_names.append(name)
        self.ee_layer_dict[name] = {
            "ee_object": ee_object,
            "ee_layer": tile_layer,
            "vis_params": vis_params,
        }

        self.add_layer(tile_layer)

        if isinstance(ee_object, ee.Image) or isinstance(ee_object, ee.ImageCollection):
            self.ee_raster_layers.append(ee_object)
            self.ee_raster_layer_names.append(name)
            if self.plot_dropdown_widget is not None:
                self.plot_dropdown_widget.options = list(self.ee_raster_layer_names)
        elif (
            isinstance(ee_object, ee.Geometry)
            or isinstance(ee_object, ee.Feature)
            or isinstance(ee_object, ee.FeatureCollection)
        ):
            self.ee_vector_layers.append(ee_object)
            self.ee_vector_layer_names.append(name)

    addLayer = add_ee_layer

    def find_layer(self, name):
        """Finds layer by name. Source: Dr. Qiusheng Wu: https://github.com/giswqs/geemap/blob/master/geemap/geemap.py
        Args:
            name (str): Name of the layer to find.
        Returns:
            object: ipyleaflet layer object.
        """
        layers = self.layers

        for layer in layers:
            if layer.name == name:
                return layer

        return None

    def find_layer_index(self, name):
        """Finds layer index by name. Source: Dr. Qiusheng Wu: https://github.com/giswqs/geemap/blob/master/geemap/geemap.py
        Args:
            name (str): Name of the layer to find.
        Returns:
            int: Index of the layer with the specified name
        """
        layers = self.layers

        for index, layer in enumerate(layers):
            if layer.name == name:
                return index

        return -1

    def toolbar_reset(self):
        """Reset the toolbar so that no tool is selected. Source: Dr. Qiusheng Wu: https://github.com/giswqs/geemap/blob/master/geemap/geemap.py"""
        toolbar_grid = self.toolbar
        for tool in toolbar_grid.children:
            tool.value = False

# Credit: Dr. Qiusheng Wu
def shp_to_geojson(in_shp, out_geojson=None):
    """Converts a shapefile to GeoJSON.

    Args:
        in_shp (str): The file path to the input shapefile.
        out_geojson (str, optional): The file path for the output GeoJSON. Defaults to None.

    Raises:
        FileNotFoundError: If the provided file path does not exist.
    """
    import json
    import shapefile

    in_shp = os.path.abspath(in_shp)

    if not os.path.exists(in_shp):
        raise FileNotFoundError("The provided shapefile could not be found.")

    sf = shapefile.Reader(in_shp)
    geojson = sf.__geo_interface__

    if out_geojson is None:
        return geojson
    else:
        out_geojson = os.path.abspath(out_geojson)
        out_dir = os.path.dirname(out_geojson)
        if not os.path.exists(out_dir):
            os.makedirs(out_dir)
        with open(out_geojson, "w") as f:
            f.write(json.dumps(geojson))    

def gmapjson_to_geojson(in_gmapjson, out_gmapgeojson=None):
    """Converts a Google Map Location History JSON to GeoJSON.

    Args:
        in_gmapjson (str): The file path to the input JSON.
        out_gmapgeojson (str, optional): The file path for the output GeoJSON. Defaults to None.

    Raises:
        FileNotFoundError: If the provided file path does not exist.
    """

    import json
    import datetime
    
    in_gmapjson = os.path.abspath(in_gmapjson)

    if not os.path.exists(in_gmapjson):
       raise FileNotFoundError("The provided json could not be found.")

    with open(in_gmapjson) as f:
        data = json.load(f)
    
    for item in data["locations"]:
        item["latitudeE7"] = item["latitudeE7"] * 0.0000001
        item["longitudeE7"] = item["longitudeE7"] * 0.0000001
        item["timestampMs"] = datetime.datetime.fromtimestamp(int(item["timestampMs"])/1000.0)
        item["timestampMs"] = item["timestampMs"].strftime('%Y-%m-%d %H:%M:%S') 
    
    data_intermediate = data["locations"]

    geojson = {
        "type": "FeatureCollection",
        "features": [
        {
            "type": "Feature",
            "geometry" : {
                "type": "Point",
                "coordinates": [d["longitudeE7"], d["latitudeE7"]],
                },
            "properties" : d,
        } for d in data_intermediate]
    }

    if out_gmapgeojson is None:
        return geojson
    else:
        out_gmapgeojson = os.path.abspath(out_gmapgeojson)
        out_dir = os.path.dirname(out_gmapgeojson)
        if not os.path.exists(out_dir):
            os.makedirs(out_dir)
        with open(out_gmapgeojson, "w") as f:
            f.write(json.dumps(geojson)) 


def gdf_to_geojson(gdf, out_geojson=None):
    """Convert a geopandas.GeoDataFrame into a GeoJSON.
    
    Args:
        gdf (gpd.GeoDataFrame): A GeoPandas GeoDataFrame.
        out_geojson (str): File path to the output GeoJSON.

    Returns:
        GeoJSON: GeoJSON of a converted pandas DataFrame
    """
    
 #   if not isinstance(gdf, gpd.GeoDataFrame):
 #       err_str = "\n\nThe gdf argument must be an instance of a gpd.GeoDataFrame"
 #       raise TypeError(err_str)
    
    geojson = gdf.__geo_interface__

    if out_geojson is None:
        return geojson
    else:
        out_geojson = os.path.abspath(out_geojson)
        out_dir = os.path.dirname(out_geojson)
        if not os.path.exists(out_dir):
            os.makedirs(out_dir)
        with open(out_geojson, "w") as f:
            f.write(json.dumps(geojson)) 


# def df_to_geojson(df, properties=df.columns.tolist(), vector_geom='Point', coords=["latitude","longitude"], out_geojson=None):
#     """Convert a pandas.DataFrame or geopandas.GeoDataFrame or hagerstrand.ExtendedDataFrame into a GeoJSON.
#        Source Inspiration: Geoff Boeing -- https://geoffboeing.com/2015/10/exporting-python-data-geojson/
    
#     Args:
#         df (pd.DataFrame): A pandas DataFrame.
#         properties ([type]): Columns to store as feature attributes.
#         vector_geom (str, optional): Shapely Geometry type. Defaults to 'Point'.
#         coords (list, optional): List containing column(s) of coordinates.

#     Returns:
#         GeoJSON: GeoJSON of a converted pandas DataFrame
#     """
    
#     if isinstance(df, pd.DataFrame):
#         df = gpd.GeoDataFrame(df.copy())

#    geojson = {'type':'FeatureCollection', 'features':[]}

#    if vector_geom == "Point":
#        for _, row in df.iterrows():
#            feature = {'type':'Feature',
#                    'properties':{},
#                    'geometry':{'type':vector_geom,
#                                'coordinates':[]}}
#            feature['geometry']['coordinates'] = [row[1],row[0]]
#            for prop in properties:
#                feature['properties'][prop] = row[prop]
#            geojson['features'].append(feature)
#    else:
#     geojson = df.__geo_interface__

#     if out_geojson is None:
#         return geojson
#     else:
#         out_geojson = os.path.abspath(out_geojson)
#         out_dir = os.path.dirname(out_geojson)
#         if not os.path.exists(out_dir):
#             os.makedirs(out_dir)
#         with open(out_geojson, "w") as f:
#             f.write(json.dumps(geojson)) 


def ee_tile_layer(
    ee_object, vis_params={}, name="Layer untitled", shown=True, opacity=1.0
):
    """Converts and Earth Engine layer to ipyleaflet TileLayer.
    Args:
        ee_object (Collection|Feature|Image|MapId): The object to add to the map.
        vis_params (dict, optional): The visualization parameters. Defaults to {}.
        name (str, optional): The name of the layer. Defaults to 'Layer untitled'.
        shown (bool, optional): A flag indicating whether the layer should be on by default. Defaults to True.
        opacity (float, optional): The layer's opacity represented as a number between 0 and 1. Defaults to 1.
    """

    image = None

    if (
        not isinstance(ee_object, ee.Image)
        and not isinstance(ee_object, ee.ImageCollection)
        and not isinstance(ee_object, ee.FeatureCollection)
        and not isinstance(ee_object, ee.Feature)
        and not isinstance(ee_object, ee.Geometry)
    ):
        err_str = "\n\nThe image argument in 'addLayer' function must be an instance of one of ee.Image, ee.Geometry, ee.Feature or ee.FeatureCollection."
        raise AttributeError(err_str)

    if (
        isinstance(ee_object, ee.geometry.Geometry)
        or isinstance(ee_object, ee.feature.Feature)
        or isinstance(ee_object, ee.featurecollection.FeatureCollection)
    ):
        features = ee.FeatureCollection(ee_object)

        width = 2

        if "width" in vis_params:
            width = vis_params["width"]

        color = "000000"

        if "color" in vis_params:
            color = vis_params["color"]

        image_fill = features.style(**{"fillColor": color}).updateMask(
            ee.Image.constant(0.5)
        )
        image_outline = features.style(
            **{"color": color, "fillColor": "00000000", "width": width}
        )

        image = image_fill.blend(image_outline)
    elif isinstance(ee_object, ee.image.Image):
        image = ee_object
    elif isinstance(ee_object, ee.imagecollection.ImageCollection):
        image = ee_object.mosaic()

    map_id_dict = ee.Image(image).getMapId(vis_params)
    tile_layer = TileLayer(
        url=map_id_dict["tile_fetcher"].url_format,
        attribution="Google Earth Engine",
        name=name,
        opacity=opacity,
        visible=shown,
    )
    return tile_layer


def poly_centroid(in_poly_shp, orig_epsg=6576, change_crs=False, epsg=6576, out_shp=None):
    """Calculates the centroid of all polygons in a file or DataFrame or GeoDataFrame.
    
    Args:
        in_poly_shp (str|pd.DataFrame|gpd.GeoDataFrame): The file path or pd.DataFrame or gpd.GeoDataFrame to the input polygon layer.
        orig_epsg (int, optional): The epsg code for the coordinate system of your input polygon layer. Defaults to 6576 (NAD83(2011) Tennessee State Plane).)
        change_crs (bool, optional): A flag indicating whether the coordinate system should be changed. Defaults to False.
        epsg (int, optional): The epsg code for the coordinate system to be changed. Defaults to 6576 (NAD83(2011) Tennessee State Plane).
        out_shp (str, optional): The filepath for the output shapefile. Defaults to None; doesn't save.

    Raises:
        FileNotFoundError: If the provided file path does not exist.
        TypeError: If the input geojson is not a str or dict.
    """           
    import geopandas as gpd  
    
    if isinstance(in_poly_shp, str):
        
        if not os.path.exists(in_poly_shp):
            raise FileNotFoundError("The provided shapefile could not be found.")        
        
        orig_crs = "epsg:" + str(orig_epsg)
        poly = gpd.read_file(in_poly_shp, crs=orig_crs)
        points = poly.copy()
        
    elif isinstance(in_poly_shp, gpd.GeoDataFrame):
        points = in_poly_shp.copy()
    
    elif isinstance(in_poly_shp, pd.DataFrame):
        points = in_poly_shp.copy()
    
    else:
        raise TypeError("The input polygon layer must be a type of str, pandas.DataFrame, or geopandas.GeoDataFrame.")
    
    if change_crs:
        epsg_str = "epsg:" + str(epsg)
        poly = poly.to_crs(epsg_str)
    
    points.crs = poly.crs
    print("The coordinate system for the centroids is: " + str(points.crs))
    print("The coordinate system for the polygons is: " + str(poly.crs))
    points.geometry = points['geometry'].centroid
    
    if out_shp is not None:
        points.to_file(out_shp)

    return points

def csv_to_gdf(in_csv, index_col=None, latloncols=["latitude","longitude"], in_epsg=4326, out_epsg=6576, out_shp=None):
    """Converts a spreadsheet with latitude and longitude values to a geopandas.GeoDataFrame

    Args:
        in_csv (str): The file path to the input csv.
        index_col (int, optional): The index column. Defaults to None.
        latloncols (list, optional): The columns of y and x (e.g. lat, lon) as a list. Defaults to ["latitude","longitude"].
        in_epsg (int, optional): The EPSG code of the csv. Defaults to 4326 (WGS84).
        out_epsg (int, optional): The EPSG code of the output, if desired. Defaults to 6576 (NAD83(2011) State Plane Tennessee).
        out_shp (str, optional): The filepath for the output shapefile. Defaults to None; doesn't save.

    Returns:
        poi (gpd.GeoDataFrame): The output geopandas.GeoDataFrame.
    """
    import pandas as pd
    import geopandas as gpd
    from shapely.geometry import Point
    
    if index_col is not None:
        poi = pd.read_csv(in_csv, index_col=index_col)
    else:
        poi = pd.read_csv(in_csv)

    poi["ycoord"] = poi[latloncols[0]].astype(float)
    poi["xcoord"] = poi[latloncols[1]].astype(float)
    epsg_i = "epsg:" + str(in_epsg)
    poi = gpd.GeoDataFrame(
        poi,
        geometry = [Point(x,y) for x, y in zip(poi.xcoord, poi.ycoord)],
        crs = epsg_i
    )
    
    epsg_o = "epsg:" + str(out_epsg)
    poi = poi.to_crs({"init":epsg_o})
    
    if out_shp is not None:
        poi.to_file(out_shp)  

    return poi

def get_nearest(src_points, candidates, metric='euclidean', k_neighbors=1):
    """Find nearest neighbors for all source points from a set of candidate points

    Args:
        src_points (np.array): Numpy array containing x and y coordinates of each source point.
        candidates (np.array): Numpy array containing x and y coordinates of each potential destination point.
        metric (str, optional): The measure of distance to use to calculate nearest neighbors. Defaults to 'euclidean'.
        k_neighbors (int, optional): Number of nearest neighbors to find. Defaults to 1.

    Returns:
        closest (np.array): Numpy array of indices of closest candidate to each source point.
        closest_dict (np.array): Numpy array of closest distances between source point and closest candidate point.
    """
    from sklearn.neighbors import BallTree
    import numpy as np

    # Create tree from the candidate points
    tree = BallTree(candidates, leaf_size=15, metric=metric)

    # Find closest points and distances
    distances, indices = tree.query(src_points, k=k_neighbors)

    # Transpose to get distances and indices into arrays
    distances = distances.transpose()
    indices = indices.transpose()

    # Get closest indices and distances (i.e. array at index 0)
    # note: for the second closest points, you would take index 1, etc.
    closest = indices[0]
    closest_dist = distances[0]

    # Return indices and distances
    return (closest, closest_dist)


def nearest_neighbor(left_gdf, right_gdf, metric="euclidean", k_neighbors=1, return_dist=False):
    """Find the nearest neighbor in right_gdf for each point in left_gdf and return the distance between them.

    Args:
        left_gdf (geopandas.GeoDataFrame): GeoDataFrame containing origin locations. This assumes your x and y coordinates are in feet.
        right_gdf (geopandas.GeoDataFrame): GeoDataFrame containing potential destination locations. This assumes your x and y coordinates are in feet.
        metric (str, optional): The measure of distance to use to calculate nearest neighbors. Defaults to "euclidean".
        k_neighbors (int, optional): Number of nearest neighbors to find. Defaults to 1.
        return_dist (bool, optional): A flag indicating whether distances should be returned. Defaults to False.

    Returns:
        closest_points (geopandas.GeoDataFrame): Closest destination locations from right_gdf to each origin location in left_gdf.
    """

    import numpy as np

    #For each point in left_gdf, find closest point in right GeoDataFrame and return them.

    left_geom_col = left_gdf.geometry.name
    right_geom_col = right_gdf.geometry.name

    # Ensure that index in right gdf is formed of sequential numbers
    right = right_gdf.copy().reset_index(drop=True)
    
    if metric == "euclidean":
        # Parse coordinates from points and insert them into a numpy array as FEET
        left_measure = np.array(left_gdf[left_geom_col].apply(lambda geom: (geom.x, geom.y)).to_list())
        right_measure = np.array(right[right_geom_col].apply(lambda geom: (geom.x, geom.y)).to_list())
    elif metric == "haversine":
    # Parse coordinates from points and insert them into a numpy array as RADIANS
        left_radians = np.array(left_gdf[left_geom_col].apply(lambda geom: (geom.x * np.pi / 180, geom.y * np.pi / 180)).to_list())
        right_radians = np.array(right[right_geom_col].apply(lambda geom: (geom.x * np.pi / 180, geom.y * np.pi / 180)).to_list())

    
    # Find the nearest points
    # -----------------------
    # closest ==> index in right_gdf that corresponds to the closest point
    # dist ==> distance between the nearest neighbors (in meters)

    closest, dist = get_nearest(src_points=left_measure, candidates=right_measure, metric=metric, k_neighbors=k_neighbors)

    # Return points from right GeoDataFrame that are closest to points in left GeoDataFrame
    closest_points = right.loc[closest]

    # Ensure that the index corresponds the one in left_gdf
    closest_points = closest_points.reset_index(drop=True)

    # Add distance if requested
    if return_dist:
        if metric == "euclidean":
            closest_points['distance'] = dist
        elif metric == "haversine":
            # Convert to miles from radians
            earth_radius = 3,958.7558657  # miles
            closest_points['distance'] = dist * earth_radius

    # We should have exactly the same number of closest_points as we have origin locations
    print("# of closest points:", len(closest_points), '==', "# of origin locations:", len(left_gdf))

    # Rename the geometry of closest stores gdf so that we can easily identify it
    closest_points = closest_points.rename(columns={'geometry': 'closest_poi_geom'})
    
    # Merge the datasets by index (for this, it is good to use '.join()' -function)
    left_gdf = left_gdf.join(closest_points)
    
    return closest_points


import pandas as pd
import json
import ipywidgets as widgets
import hagerstrand as hs


class ExtendedDataFrame(pd.DataFrame):
    """This ExtendedDataFrame class inherits the pandas DataFrame class

    Args:
        pd (pd.DataFrame()): A pandas DataFrame
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def _constructor(self):
        return ExtendedDataFrame

    def deduplicate(self, columns=None):
        """Drops duplicate records and resets the index of the ExtendedDataFrame

        Args:
            columns (list, optional): Columns for which to identify dupulicate records. Defaults to None.

        Returns:
            ExtendedDataFrame: A de-deuplicated ExtendedDataFrame with the index reset.
        """
        if columns == None:
            df = self.drop_duplicates()
            df = df.reset_index(drop=True)
            return df
        else:
            df = self.drop_duplicates(subset=columns)
            df = df.reset_index(drop=True)
            return df

    def jsoncol_newdf(self):
        """Unpack a JSON column and return a new ExtendedDataFrame

        Returns:
            ExtendedDataFrame: A new ExtendedDataFrame of an unpacked JSON column.
        """
        df = unpack_json(self)
        return df

    def jsoncol_merge(self):
        """Unpack a JSON column and merge to the existing DataFrame

        Returns:
            ExtendedDataFrame: A new ExtendedDataFrame that includes the original DataFrame and the unpacked JSON column.
        """
        df = unpack_json_and_merge(self)
        return df


# THE FOLLOWING FUNCTIONS ARE MODIFIED FROM https://github.com/SafeGraphInc/safegraph_py


def load_json_nan(df, json_col):
    """Load a JSON file even if there are NaNs.

    Args:
        df (pd.DataFrame): The DataFrame containing the JSON column to be loaded.
        json_col (str): The JSON column to be loaded.

    Returns:
        df (pd.Series): A pd.Series of a JSON column
    """
    return df[json_col].apply(lambda x: json.loads(x) if type(x) == str else x)


def unpack_json(
    df,
    json_column="visitor_home_cbgs",
    index_name=None,
    key_col_name=None,
    value_col_name=None,
):
    """Unpack a JSON column from a SafeGraph Patterns dataset.

    Args:
        df (pd.DataFrame): DataFrame containing the JSON column to be unpacked.
        json_column (str, optional): JSON column to be unpacked. Defaults to 'visitor_home_cbgs'.
        index_name (str, optional): Index name for new ExtendedDataFrame. Defaults to None.
        key_col_name (str, optional): Key name for new ExtendedDataFrame. Defaults to None.
        value_col_name (str, optional): Value name for new ExtendedDataFrame. Defaults to None.

    Returns:
        pd.DataFrame: DataFrame of an unpacked JSON column.
    """
    import numpy as np

    # these checks are inefficent for multithreading, but it's not a big deal
    if key_col_name is None:
        key_col_name = json_column + "_key"
    if value_col_name is None:
        value_col_name = json_column + "_value"
    if df.index.unique().shape[0] < df.shape[0]:
        raise ("ERROR -- non-unique index found")
    df = df.copy()
    df[json_column + "_dict"] = load_json_nan(df, json_column)
    all_sgpid_cbg_data = []  # each cbg data point will be one element in this list
    if index_name is None:
        for index, row in df.iterrows():
            if row[json_column] == "" or pd.isnull(row[json_column]):
                next
            else:
                this_sgpid_cbg_data = [
                    {"orig_index": index, key_col_name: key, value_col_name: value}
                    for key, value in row[json_column + "_dict"].items()
                ]
                all_sgpid_cbg_data = all_sgpid_cbg_data + this_sgpid_cbg_data
    else:
        for index, row in df.iterrows():
            if row[json_column] == "" or pd.isnull(row[json_column]):
                next
            else:
                temp = row[index_name]
                this_sgpid_cbg_data = [
                    {
                        "orig_index": index,
                        index_name: temp,
                        key_col_name: key,
                        value_col_name: value,
                    }
                    for key, value in row[json_column + "_dict"].items()
                ]
                all_sgpid_cbg_data = all_sgpid_cbg_data + this_sgpid_cbg_data

    all_sgpid_cbg_data = pd.DataFrame(all_sgpid_cbg_data)
    all_sgpid_cbg_data.set_index("orig_index", inplace=True)
    return all_sgpid_cbg_data


def unpack_json_and_merge(
    df,
    json_column="visitor_home_cbgs",
    key_col_name="visitor_home_cbg",
    value_col_name="cbg_visitor_name",
    keep_index=False,
):
    """Unpack a JSON column from a SafeGraph Patterns dataset.

    Args:
        df (pd.DataFrame): DataFrame containing the JSON column to be unpacked
        json_column (str, optional): JSON column to be unpacked. Defaults to 'visitor_home_cbgs'.
        key_col_name (str, optional): Key name for new ExtendedDataFrame. Defaults to 'visitor_home_cbg'.
        value_col_name (str, optional): Value name for new ExtendedDataFrame. Defaults to 'cbg_visitor_name'.
        keep_index (bool, optional): Keep or do not keep the original index. Defaults to False.

    Returns:
        pd.DataFrame: DataFrame containing the original DataFrame and the unpacked JSON column as additional columns.
    """
    if keep_index:
        df["index_original"] = df.index
    df.reset_index(drop=True, inplace=True)  # Every row must have a unique index
    df_exp = unpack_json(
        df,
        json_column=json_column,
        key_col_name=key_col_name,
        value_col_name=value_col_name,
    )
    df = df.merge(df_exp, left_index=True, right_index=True).reset_index(drop=True)
    return df


def unique_sorted_values_plus_ALL(array):
    """Obtain a sorted array of all unique values in an array, including an additional value of 'ALL' to denote all values.

    Args:
        array (list|pd.Series|np.array): An array of values.

    Returns:
        list: A sorted array of unique values including an additional value of 'ALL'.
    """
    unique = array.unique().tolist()
    unique.sort()
    unique.insert(0, "ALL")
    return unique


def unique_sorted_columns_plus_ALL(gj):
    """Obtain a sorted array of all columns in a GeoJSON, including an additional value of 'ALL' to denote all values.

    Args:
        gj (GeoJSON): A GeoJSON.

    Returns:
        list: A sorted array of unique column names including an additional value of 'ALL'.
    """
    if isinstance(gj, json):
        array = gpd.GeoDataFrame.from_features(gj)
        unique = array.unique().tolist()
        unique.sort()
        unique.insert(0, "ALL")
        return unique

    else:
        raise TypeError("The provided argument for gj must be of type json.")


# Come back to this: https://stackoverflow.com/questions/42562895/return-dataframe-using-ipywidgets-button



"""Source Credit: Dr. Qiusheng Wu -- https://github.com/giswqs/geodemo/blob/master/geodemo/utils.py"""

def random_string(string_length=3, use_seed=False):
    """Generates a random string of fixed length.
    Args:
        string_length (int, optional): Fixed length. Defaults to 3.
    Returns:
        str: A random string
    """
    import random
    import string

    if use_seed:
        random.seed(1001)
    letters = string.ascii_lowercase
    return "".join(random.choice(letters) for i in range(string_length))


"""Module for basemaps. Each basemap is defined as item in the ee_basemaps dictionary. For example, to access Google basemaps, use the following:
ee_basemaps['ROADMAP'], ee_basemaps['SATELLITE'], ee_basemaps['HYBRID'].
More WMS basemaps can be found at the following websites:
1. USGS National Map: https://viewer.nationalmap.gov/services/
2. MRLC NLCD Land Cover data: https://viewer.nationalmap.gov/services/
3. FWS NWI Wetlands data: https://www.fws.gov/wetlands/Data/Web-Map-Services.html
Source/Credit: Dr. Qiusheng Wu - https://github.com/giswqs/geemap/blob/master/geemap/basemaps.py
"""

from box import Box
from ipyleaflet import TileLayer, WMSLayer, basemap_to_tiles
import ipyleaflet.basemaps as ipybasemaps


_ee_basemaps = {
    "ROADMAP": TileLayer(
        url="https://mt1.google.com/vt/lyrs=m&x={x}&y={y}&z={z}",
        attribution="Google",
        name="Google Maps",
    ),
    "SATELLITE": TileLayer(
        url="https://mt1.google.com/vt/lyrs=s&x={x}&y={y}&z={z}",
        attribution="Google",
        name="Google Satellite",
    ),
    "TERRAIN": TileLayer(
        url="https://mt1.google.com/vt/lyrs=p&x={x}&y={y}&z={z}",
        attribution="Google",
        name="Google Terrain",
    ),
    "HYBRID": TileLayer(
        url="https://mt1.google.com/vt/lyrs=y&x={x}&y={y}&z={z}",
        attribution="Google",
        name="Google Satellite",
    ),
    "ESRI": TileLayer(
        url="https://server.arcgisonline.com/ArcGIS/rest/services/World_Imagery/MapServer/tile/{z}/{y}/{x}",
        attribution="Esri",
        name="Esri Satellite",
    ),
    "Esri Ocean": TileLayer(
        url="https://services.arcgisonline.com/ArcGIS/rest/services/Ocean/World_Ocean_Base/MapServer/tile/{z}/{y}/{x}",
        attribution="Esri",
        name="Esri Ocean",
    ),
    "Esri Satellite": TileLayer(
        url="https://server.arcgisonline.com/ArcGIS/rest/services/World_Imagery/MapServer/tile/{z}/{y}/{x}",
        attribution="Esri",
        name="Esri Satellite",
    ),
    "Esri Standard": TileLayer(
        url="https://server.arcgisonline.com/ArcGIS/rest/services/World_Street_Map/MapServer/tile/{z}/{y}/{x}",
        attribution="Esri",
        name="Esri Standard",
    ),
    "Esri Terrain": TileLayer(
        url="https://server.arcgisonline.com/ArcGIS/rest/services/World_Terrain_Base/MapServer/tile/{z}/{y}/{x}",
        attribution="Esri",
        name="Esri Terrain",
    ),
    "Esri Transportation": TileLayer(
        url="https://server.arcgisonline.com/ArcGIS/rest/services/Reference/World_Transportation/MapServer/tile/{z}/{y}/{x}",
        attribution="Esri",
        name="Esri Transportation",
    ),
    "Esri Topo World": TileLayer(
        url="https://services.arcgisonline.com/ArcGIS/rest/services/World_Topo_Map/MapServer/tile/{z}/{y}/{x}",
        attribution="Esri",
        name="Esri Topo World",
    ),
    "Esri National Geographic": TileLayer(
        url="http://services.arcgisonline.com/ArcGIS/rest/services/NatGeo_World_Map/MapServer/tile/{z}/{y}/{x}",
        attribution="Esri",
        name="Esri National Geographic",
    ),
    "Esri Shaded Relief": TileLayer(
        url="https://services.arcgisonline.com/arcgis/rest/services/World_Shaded_Relief/MapServer/tile/{z}/{y}/{x}",
        attribution="Esri",
        name="Esri Shaded Relief",
    ),
    "Esri Physical Map": TileLayer(
        url="https://services.arcgisonline.com/arcgis/rest/services/World_Physical_Map/MapServer/tile/{z}/{y}/{x}",
        attribution="Esri",
        name="Esri Physical Map",
    ),
    "FWS NWI Wetlands": WMSLayer(
        url="https://www.fws.gov/wetlands/arcgis/services/Wetlands/MapServer/WMSServer?",
        layers="1",
        name="FWS NWI Wetlands",
        attribution="FWS",
        format="image/png",
        transparent=True,
    ),
    "FWS NWI Wetlands Raster": WMSLayer(
        url="https://www.fws.gov/wetlands/arcgis/services/Wetlands_Raster/ImageServer/WMSServer?",
        layers="0",
        name="FWS NWI Wetlands Raster",
        attribution="FWS",
        format="image/png",
        transparent=True,
    ),
    "Google Maps": TileLayer(
        url="https://mt1.google.com/vt/lyrs=m&x={x}&y={y}&z={z}",
        attribution="Google",
        name="Google Maps",
    ),
    "Google Satellite": TileLayer(
        url="https://mt1.google.com/vt/lyrs=s&x={x}&y={y}&z={z}",
        attribution="Google",
        name="Google Satellite",
    ),
    "Google Terrain": TileLayer(
        url="https://mt1.google.com/vt/lyrs=p&x={x}&y={y}&z={z}",
        attribution="Google",
        name="Google Terrain",
    ),
    "Google Satellite Hybrid": TileLayer(
        url="https://mt1.google.com/vt/lyrs=y&x={x}&y={y}&z={z}",
        attribution="Google",
        name="Google Satellite",
    ),
    "NLCD 2016 CONUS Land Cover": WMSLayer(
        url="https://www.mrlc.gov/geoserver/mrlc_display/NLCD_2016_Land_Cover_L48/wms?",
        layers="NLCD_2016_Land_Cover_L48",
        name="NLCD 2016 CONUS Land Cover",
        attribution="MRLC",
        format="image/png",
        transparent=True,
    ),
    "NLCD 2013 CONUS Land Cover": WMSLayer(
        url="https://www.mrlc.gov/geoserver/mrlc_display/NLCD_2013_Land_Cover_L48/wms?",
        layers="NLCD_2013_Land_Cover_L48",
        name="NLCD 2013 CONUS Land Cover",
        attribution="MRLC",
        format="image/png",
        transparent=True,
    ),
    "NLCD 2011 CONUS Land Cover": WMSLayer(
        url="https://www.mrlc.gov/geoserver/mrlc_display/NLCD_2011_Land_Cover_L48/wms?",
        layers="NLCD_2011_Land_Cover_L48",
        name="NLCD 2011 CONUS Land Cover",
        attribution="MRLC",
        format="image/png",
        transparent=True,
    ),
    "NLCD 2008 CONUS Land Cover": WMSLayer(
        url="https://www.mrlc.gov/geoserver/mrlc_display/NLCD_2008_Land_Cover_L48/wms?",
        layers="NLCD_2008_Land_Cover_L48",
        name="NLCD 2008 CONUS Land Cover",
        attribution="MRLC",
        format="image/png",
        transparent=True,
    ),
    "NLCD 2006 CONUS Land Cover": WMSLayer(
        url="https://www.mrlc.gov/geoserver/mrlc_display/NLCD_2006_Land_Cover_L48/wms?",
        layers="NLCD_2006_Land_Cover_L48",
        name="NLCD 2006 CONUS Land Cover",
        attribution="MRLC",
        format="image/png",
        transparent=True,
    ),
    "NLCD 2004 CONUS Land Cover": WMSLayer(
        url="https://www.mrlc.gov/geoserver/mrlc_display/NLCD_2004_Land_Cover_L48/wms?",
        layers="NLCD_2004_Land_Cover_L48",
        name="NLCD 2004 CONUS Land Cover",
        attribution="MRLC",
        format="image/png",
        transparent=True,
    ),
    "NLCD 2001 CONUS Land Cover": WMSLayer(
        url="https://www.mrlc.gov/geoserver/mrlc_display/NLCD_2001_Land_Cover_L48/wms?",
        layers="NLCD_2001_Land_Cover_L48",
        name="NLCD 2001 CONUS Land Cover",
        attribution="MRLC",
        format="image/png",
        transparent=True,
    ),
    "USGS NAIP Imagery": WMSLayer(
        url="https://services.nationalmap.gov/arcgis/services/USGSNAIPImagery/ImageServer/WMSServer?",
        layers="0",
        name="USGS NAIP Imagery",
        attribution="USGS",
        format="image/png",
        transparent=True,
    ),
    "USGS Hydrography": WMSLayer(
        url="https://basemap.nationalmap.gov/arcgis/services/USGSHydroCached/MapServer/WMSServer?",
        layers="0",
        name="USGS Hydrography",
        attribution="USGS",
        format="image/png",
        transparent=True,
    ),
    "USGS 3DEP Elevation": WMSLayer(
        url="https://elevation.nationalmap.gov/arcgis/services/3DEPElevation/ImageServer/WMSServer?",
        layers="3DEPElevation:None",
        name="USGS 3DEP Elevation",
        attribution="USGS",
        format="image/png",
        transparent=True,
    ),
}

# Adds ipyleaflet basemaps
for item in ipybasemaps.values():
    try:
        name = item["name"]
        basemap = "ipybasemaps.{}".format(name)
        _ee_basemaps[name] = basemap_to_tiles(eval(basemap))
    except Exception:
        for sub_item in item:
            name = item[sub_item]["name"]
            basemap = "ipybasemaps.{}".format(name)
            basemap = basemap.replace("Mids", "Modis")
            _ee_basemaps[name] = basemap_to_tiles(eval(basemap))

basemap_tiles = Box(_ee_basemaps, frozen_box=True)
basemaps = Box(
    dict(zip(list(_ee_basemaps.keys()), list(_ee_basemaps.keys()))), frozen_box=True)


"""A module for interactive toolbars with ipywidgets.
Source: Dr. Qiusheng Wu -- https://github.com/giswqs/geodemo/blob/master/geodemo/toolbar.py
"""

import os
import ipywidgets as widgets
import ipyleaflet
from ipyleaflet import WidgetControl
from ipyfilechooser import FileChooser
from IPython.display import display
from .dataprocess import unique_sorted_values_plus_ALL

def change_basemap(m):
    """Widget for change basemaps. Source: Dr. Qiusheng Wu -- https://github.com/giswqs/geemap/blob/master/geemap/toolbar.py
    Args:
        m (object): hagerstrand.Map()
    """
    from .basemaps import _ee_basemaps

    dropdown = widgets.Dropdown(
        options=list(_ee_basemaps.keys()),
        value="ROADMAP",
        layout=widgets.Layout(width="200px"),
        description="Basemaps",
    )

    close_btn = widgets.Button(
        icon="times",
        tooltip="Close the basemap widget",
        button_style="primary",
        layout=widgets.Layout(width="32px"),
    )

    basemap_widget = widgets.HBox([dropdown, close_btn])

    def on_click(change):
        basemap_name = change["new"]

        if len(m.layers) == 1:
            old_basemap = m.layers[0]
        else:
            old_basemap = m.layers[1]
        m.substitute_layer(old_basemap, _ee_basemaps[basemap_name])

    dropdown.observe(on_click, "value")

    def close_click(change):
        m.toolbar_reset()
        if m.basemap_ctrl is not None and m.basemap_ctrl in m.controls:
            m.remove_control(m.basemap_ctrl)
        basemap_widget.close()

    close_btn.on_click(close_click)

    basemap_control = WidgetControl(widget=basemap_widget, position="topright")
    m.add_control(basemap_control)
    m.basemap_ctrl = basemap_control


def filter_df_widget(df, field):
    """Widget for filtering a DataFrame

    Args:
        df (pd.DataFrame): A DataFrame to filter
        field (str): A field within the DataFrame for the filter criterion
    """
    
    dropdown_field = widgets.Dropdown(
        options = unique_sorted_values_plus_ALL(df[field]),
        value="ALL",
        layout=widgets.Layout(width="200px"),
        description=field
    )

    close_dropdown_field = widgets.Button(
        icon="times",
        tooltip="Close the filter widget",
        button_style="primary",
        layout=widgets.Layout(width="32px")
    )

    filter_widget = widgets.HBox([dropdown_field, close_dropdown_field])

    out_field = widgets.Output()

    def on_click_dropdown(change):
        with out_field:
            out_field.clear_output()
            if (change.new == 'ALL'):
                display(df)
            else:
                display(df[df[field] == change.new])

    dropdown_field.observe(on_click_dropdown, names="value")

    def close_click_dropdown(change):
        filter_widget.close()
    
    close_dropdown_field.on_click(close_click_dropdown)
    
    display(filter_widget)
    display(out_field)


def main_toolbar(m):

    padding = "0px 0px 0px 5px"  # upper, right, bottom, left

    toolbar_button = widgets.ToggleButton(
        value=False,
        tooltip="Toolbar",
        icon="wrench",
        layout=widgets.Layout(width="28px", height="28px", padding=padding),
    )

    close_button = widgets.ToggleButton(
        value=False,
        tooltip="Close the tool",
        icon="times",
        button_style="primary",
        layout=widgets.Layout(height="28px", width="28px", padding=padding),
    )    

    toolbar = widgets.HBox([toolbar_button])

    def close_click(change):
        if change["new"]:
            toolbar_button.close()
            close_button.close()
            toolbar.close()
            
    close_button.observe(close_click, "value")

    rows = 2
    cols = 2
    grid = widgets.GridspecLayout(rows, cols, grid_gap="0px", layout=widgets.Layout(width="62px"))

    icons = ["folder-open", "map", "gears", "filter"]

    for i in range(rows):
        for j in range(cols):
            grid[i, j] = widgets.Button(description="", button_style="primary", icon=icons[i*rows+j], 
                                        layout=widgets.Layout(width="28px", padding="0px"))

    toolbar = widgets.VBox([toolbar_button])

    def toolbar_click(change):
        if change["new"]:
            toolbar.children = [widgets.HBox([close_button, toolbar_button]), grid]
        else:
            toolbar.children = [toolbar_button]
        
    toolbar_button.observe(toolbar_click, "value")

    toolbar_ctrl = WidgetControl(widget=toolbar, position="topright")

    m.add_control(toolbar_ctrl)

    output = widgets.Output()
    output_ctrl = WidgetControl(widget=output, position="topright")

    buttons = widgets.ToggleButtons(
        value=None,
        options=["Apply", "Reset", "Close"],
        tooltips=["Apply", "Reset", "Close"],
        button_style="primary",
    )
    buttons.style.button_width = "80px"

    data_dir = os.path.abspath('./data')

    # File Chooser Widget
    fc = FileChooser(data_dir)
    fc.use_dir_icons = True
    fc.filter_pattern = ['*.shp', '*.geojson', '*.json']

    filechooser_widget = widgets.VBox([fc, buttons])

    def button_click(change):
        if change["new"] == "Apply" and fc.selected is not None:
            if fc.selected.endswith(".shp"):
                m.add_shapefile(fc.selected, layer_name="Shapefile")
            elif fc.selected.endswith(".geojson"):
                m.add_geojson(fc.selected, layer_name="GeoJSON")
            elif fc.selected.endswith(".json"):
                m.add_geojson(fc.selected, layer_name="GeoJSON")
        elif change["new"] == "Reset":
            fc.reset()
        elif change["new"] == "Close":
            fc.reset()
            m.remove_control(output_ctrl)
    buttons.observe(button_click, "value")     

    # Basemap Widget
    from .basemaps import _ee_basemaps

    dropdown_basemap = widgets.Dropdown(
        options=list(_ee_basemaps.keys()),
        value="ROADMAP",
        layout=widgets.Layout(width="200px"),
        description="Basemaps",
    )

    close_button_basemap = widgets.Button(
        icon="times",
        tooltip="Close the basemap widget",
        button_style="primary",
        layout=widgets.Layout(width="32px"),
    )

    basemap_widget = widgets.HBox([dropdown_basemap, close_button_basemap])

    def on_click(change):
        basemap_name = change["new"]

        if len(m.layers) == 1:
            old_basemap = m.layers[0]
        else:
            old_basemap = m.layers[1]
        m.substitute_layer(old_basemap, _ee_basemaps[basemap_name])

    dropdown_basemap.observe(on_click, "value")

    def close_click(change):
        if m.basemap_ctrl is not None and m.basemap_ctrl in m.controls:
            m.remove_control(m.basemap_ctrl)
        basemap_widget.close()

    close_button_basemap.on_click(close_click)

    basemap_control = WidgetControl(widget=basemap_widget, position="topright")
    m.basemap_ctrl = basemap_control

    # Select Layer for Filtering Widget
    layers = [layer.name for layer in m.layers if not isinstance(layer, ipyleaflet.TileLayer)]

    dropdown_layer = widgets.Dropdown(
        options=layers,
        description='Layer:'
    )

    dropdown_layer_field = widgets.Dropdown(
        description='Field:'
    )

    dropdown_layer_field_value = widgets.Dropdown(
        description='Value:'
    )

    out_filter_layer = widgets.Output()

    def on_click_layer(change):
        with out_filter_layer:
            out_filter_layer.clear_output()
            dropdown_layer_field.options = list(m.find_layer(change.new).data['features'][0]['properties'].keys())

    dropdown_layer.observe(on_click_layer, names="value")

    out_filter_layer_field = widgets.Output()

    def on_click_layer_field(change):
        with out_filter_layer_field:
            out_filter_layer_field.clear_output()
            field_values = set()
            for record in m.find_layer(dropdown_layer.value).data['features']:
                field_values.add(record['properties'][dropdown_layer_field.value])
            dropdown_layer_field_value.options = sorted(field_values)

    dropdown_layer_field.observe(on_click_layer_field, names="value")


    add_filter_layer = widgets.Button(
         icon="plus",
         tooltip="Add a layer based on the filter options",
         button_color="lightgreen",
         layout=widgets.Layout(width="32px")
    )

    out_click_add_filter_layer = widgets.Output()

    def click_add_filter_layer(change):
            
        import itertools
        import copy

        data = copy.deepcopy(m.find_layer(dropdown_layer.value).data)
        filter_data = [record for record in data['features'] if record['properties'][dropdown_layer_field.value] == dropdown_layer_field_value.value]
        data['features'] = filter_data

        style = {
            "stroke": True,
            "color": "#000000",
            "weight": 2,
            "opacity": 1,
            "fill": True,
            "fillColor": "#0000ff",
            "fillOpacity": 0.4,
        }

        geojson = ipyleaflet.GeoJSON(
            data=data,
            style=style,
            name="{} - {} - {} Layer".format(
                dropdown_layer.value,
                dropdown_layer_field.value,
                dropdown_layer_field_value.value
            )
        )

        m.add_layer(geojson)

    add_filter_layer.on_click(click_add_filter_layer)

    close_dropdown_layer = widgets.Button(
         icon="times",
         tooltip="Close the filter widget",
         button_style="primary",
         layout=widgets.Layout(width="32px")
    )

    box_layout = widgets.Layout(display='flex',
                                flex_flow='row',
                                align_items='stretch',
                                width='70%')

    filter_buttons_box = widgets.Box([close_dropdown_layer, add_filter_layer])
    filter_dropdown_box = widgets.Box([dropdown_layer, dropdown_layer_field, dropdown_layer_field_value])
    filter_widget = widgets.HBox([filter_dropdown_box, filter_buttons_box])

    def close_click_dropdown(change):
        filter_widget.close()
    
    close_dropdown_layer.on_click(close_click_dropdown)
    
    filter_layer_control = WidgetControl(widget=filter_widget, position="bottomright")
    m.filter_layer_ctrl = filter_layer_control

    def tool_click(b):    
        with output:
            output.clear_output()
            if b.icon == "folder-open":
                display(filechooser_widget)
                m.add_control(output_ctrl)

            elif b.icon == "filter":
                dropdown_layer.options = [layer.name for layer in m.layers if not isinstance(layer, ipyleaflet.TileLayer)]
 #               display(layers)
                display(filter_widget)
                m.add_control(filter_layer_control)

            elif b.icon == "map":
                display(basemap_widget)
                m.add_control(basemap_control)

            elif b.icon == "gears":
                import whiteboxgui.whiteboxgui as wbt

                if hasattr(m, "whitebox") and m.whitebox is not None:
                    if m.whitebox in m.controls:
                        m.remove_control(m.whitebox)

                tools_dict = wbt.get_wbt_dict()
                wbt_toolbox = wbt.build_toolbox(
                    tools_dict, max_width="800px", max_height="500px"
                )

                wbt_control = WidgetControl(
                    widget=wbt_toolbox, position="bottomright"
                )                

                m.whitebox = wbt_control
                m.add_control(wbt_control)

#            elif b.icon == "question":

            print(f"You clicked the {b.icon} button")



    for i in range(rows):
        for j in range(cols):
            tool = grid[i, j]
            tool.on_click(tool_click)